/*
 * WTF! Wide Tile Format converter by Einar Saukas
 *
 * http://www.worldofspectrum.org
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

unsigned char *generate_wtile(unsigned char *input_data, size_t input_size, size_t *output_size) {
    unsigned char *output_data;
    unsigned char *p;
    int r, c;

    *output_size = input_size/96*72;
    output_data = (unsigned char *)malloc(*output_size);
    if (output_data) {
        p = output_data;
        while (input_size) {
            for (c=0; c < 3; c++) {
                for (r=0; r < 16; r++) {
                    p[3*r+c] = input_data[c < 2 ? 2*r+c : 2*r+48];
                }
                for (r=0; r < 8; r++) {
                    p[3*r+c+48] = input_data[c < 2 ? r+8*c+32 : r+48+32];
                }
            }
            input_data += 96;
            input_size -= 96;
            p += 72;
        }
    }
    return output_data;
}

unsigned char *generate_btile(unsigned char *input_data, size_t input_size, size_t *output_size) {
    unsigned char *output_data;
    unsigned char *p;
    int r, c;

    *output_size = input_size/72*96;
    output_data = (unsigned char *)malloc(*output_size);
    if (output_data) {
        memset(output_data, 0, *output_size);
        p = output_data;
        while (input_size) {
            for (c=0; c < 3; c++) {
                for (r=0; r < 16; r++) {
                    p[c < 2 ? 2*r+c : 2*r+48] = input_data[3*r+c];
                }
                for (r=0; r < 8; r++) {
                    p[c < 2 ? r+8*c+32 : r+48+32] = input_data[3*r+c+48];
                }
            }
            input_data += 72;
            input_size -= 72;
            p += 96;
        }
    }
    return output_data;
}

int main(int argc, char *argv[]) {
    int forced_mode = 0;
    int btile_mode = 0;
    char *input_name = NULL;
    char *output_name = NULL;
    FILE *ifp;
    FILE *ofp;
    unsigned char *input_data;
    unsigned char *output_data;
    size_t input_size;
    size_t output_size;
    size_t partial_counter;
    size_t total_counter;
    int i;

    printf("WTF! Wide Tile Format converter by Einar Saukas\n");

    /* process command-line arguments */
    for (i = 1; i < argc; i++) {
        if (!strcmp(argv[i], "-f")) {
            forced_mode = 1;
        } else if (!strcmp(argv[i], "-b")) {
            btile_mode = 1;
        } else if (input_name == NULL) {
            input_name = argv[i];
            if(strlen(input_name) >= 6 && !strcmp(input_name+strlen(input_name)-6, ".wtile")) {
                btile_mode = 1;
            }
        } else if (output_name == NULL) {
            output_name = argv[i];
        } else {
            input_name = NULL;
            break;
        }
    }

    /* validate command-line arguments */
    if (input_name == NULL) {
         fprintf(stderr, "Usage: %s [-f] input.btile [output.wtile]\n"
                         "       %s [-f] [-b] input.wtile [output.btile]\n"
                         "  -f      Force overwrite of output file\n"
                         "  -b      Convert from wtile to btile\n", argv[0], argv[0]);
         exit(1);
    }
    if (output_name == NULL) {
        output_name = (char *)malloc(strlen(input_name)+7);
        strcpy(output_name, input_name);
        if (btile_mode) {
            if(strlen(output_name) >= 6 && !strcmp(output_name+strlen(output_name)-6, ".wtile")) {
                *(output_name+strlen(output_name)-5) = 'b';
            } else {
                strcat(output_name, ".btile");
            }
        } else {
            if(strlen(output_name) >= 6 && !strcmp(output_name+strlen(output_name)-6, ".btile")) {
                *(output_name+strlen(output_name)-5) = 'w';
            } else {
                strcat(output_name, ".wtile");
            }
        }
    }

    /* open input file */
    ifp = fopen(input_name, "rb");
    if (!ifp) {
         fprintf(stderr, "Error: Cannot access input file %s\n", input_name);
         exit(1);
    }

    /* determine input size */
    fseek(ifp, 0L, SEEK_END);
    input_size = ftell(ifp);
    fseek(ifp, 0L, SEEK_SET);
    if (!input_size) {
         fprintf(stderr, "Error: Empty input file %s\n", argv[1]);
         exit(1);
    }

    /* validate input size */
    if (btile_mode ? input_size%72 : input_size%96) {
        fprintf(stderr, "Error: Invalid input file %s\n", input_name);
        exit(1);
    }

    /* allocate input buffer */
    input_data = (unsigned char *)malloc(input_size);
    if (!input_data) {
         fprintf(stderr, "Error: Insufficient memory\n");
         exit(1);
    }

    /* read input file */
    total_counter = 0;
    do {
        partial_counter = fread(input_data+total_counter, sizeof(char), input_size-total_counter, ifp);
        total_counter += partial_counter;
    } while (partial_counter > 0);

    if (total_counter != input_size) {
         fprintf(stderr, "Error: Cannot read input file %s\n", argv[1]);
         exit(1);
    }

    /* close input file */
    fclose(ifp);

    /* check output file */
    if (!forced_mode && fopen(output_name, "rb") != NULL) {
         fprintf(stderr, "Error: Already existing output file %s\n", output_name);
         exit(1);
    }

    /* create output file */
    ofp = fopen(output_name, "wb");
    if (!ofp) {
         fprintf(stderr, "Error: Cannot create output file %s\n", output_name);
         exit(1);
    }

    /* generate output file */
    output_data = btile_mode ? generate_btile(input_data, input_size, &output_size) : generate_wtile(input_data, input_size, &output_size);
    if (!output_data) {
         fprintf(stderr, "Error: Insufficient memory\n");
         exit(1);
    }

    /* write output file */
    if (fwrite(output_data, sizeof(char), output_size, ofp) != output_size) {
         fprintf(stderr, "Error: Cannot write output file %s\n", output_name);
         exit(1);
    }

    /* close output file */
    fclose(ofp);

    /* done! */                                                            
    printf("Generated %s tiles \"%s\"\n", btile_mode ? "bicolor" : "wide", output_name);

    return 0;
}
